-------------------------------------------------------------------------------
-- Broker_Currency
-- Copyright 2008+ Toadkiller of Proudmoore for the non-statistics code.
-- The statistics code is 100% ckknight
--
-- LDB display of currencies, totals and money rate for all characters on a server.
-- The statistics stuff (total money today, yesterday, last week) is 100% from FuBar_MoneyFu, credit to ckknight
-- http://www.wowace.com/projects/broker-currency/
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- Localized globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local math = _G.math
local string = _G.string
local table = _G.table

local date = _G.date
local ipairs = _G.ipairs
local pairs = _G.pairs
local select = _G.select
local time = _G.time
local tonumber = _G.tonumber
local type = _G.type
local unpack = _G.unpack

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local LibQTip = LibStub("LibQTip-1.0")

Broker_Currency = CreateFrame("frame", "Broker_CurrencyFrame")
local Broker_Currency = Broker_Currency

-- GLOBALS: Broker_CurrencyCharDB, Broker_CurrencyDB

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
-- The localization goal is to only use existing Blizzard strings and localized Title strings from the toc
local ICON_GOLD = "\124TInterface\\MoneyFrame\\UI-GoldIcon:24:24\124t"
local ICON_SILVER = "\124TInterface\\MoneyFrame\\UI-SilverIcon:24:24\124t"
local ICON_COPPER = "\124TInterface\\MoneyFrame\\UI-CopperIcon:24:24\124t"

local SETTING_ICON_STRING = "\124T%s:24:24\124t"
local DISPLAY_ICON_STRING1 = "%d\124T"
local DISPLAY_ICON_STRING2 = ":%d:%d\124t"

local fontWhite = CreateFont("Broker_CurrencyFontWhite")
local fontPlus = CreateFont("Broker_CurrencyFontPlus")
local fontMinus = CreateFont("Broker_CurrencyFontMinus")
local fontLabel = CreateFont("Broker_CurrencyFontLabel")

local PLAYER_NAME = UnitName("player")
local REALM_NAME = GetRealmName()

local sVersion = GetAddOnMetadata("Broker_Currency", "Version")

local sToday = HONOR_TODAY
local sYesterday = HONOR_YESTERDAY
local sThisWeek = ARENA_THIS_WEEK
local sLastWeek = HONOR_LASTWEEK

local sPlus = "+"
local sMinus = "-"
local sTotal = "="

local HEARTHSTONE_IDNUM = 6948

-------------------------------------------------------------------------------
-- Currencies
-------------------------------------------------------------------------------
local DALARAN_COOKING_AWARD = 81
local CHAMPIONS_SEAL = 241
local CONQUEST_POINTS = 390
local HONOR_POINTS = 392
local JUSTICE_POINTS = 395

local COIN_OF_ANCESTRY = 21100
local BREWFEST_PRIZE_TOKEN = 37829
local LOVE_TOKEN = 49927

local CURRENCY_DATA = {
	{
		currency_id = DALARAN_COOKING_AWARD,
		countFunc = function()
				    return select(2, _G.GetCurrencyInfo(DALARAN_COOKING_AWARD))
			    end
	},
	{
		currency_id = CHAMPIONS_SEAL,
		countFunc = function()
				    return select(2, _G.GetCurrencyInfo(CHAMPIONS_SEAL))
			    end
	},
	{
		currency_id = CONQUEST_POINTS,
		countFunc = function()
				    return select(2, _G.GetCurrencyInfo(CONQUEST_POINTS))
			    end
	},
	{
		currency_id = HONOR_POINTS,
		countFunc = function()
				    return select(2, _G.GetCurrencyInfo(HONOR_POINTS))
			    end
	},
	{
		currency_id = JUSTICE_POINTS,
		countFunc = function()
				    return select(2, _G.GetCurrencyInfo(JUSTICE_POINTS))
			    end
	},
	{
		item_id = COIN_OF_ANCESTRY,
		countFunc = function()
				    return _G.GetItemCount(COIN_OF_ANCESTRY, true)
			    end
	},
	{
		item_id = BREWFEST_PRIZE_TOKEN,
		countFunc = function()
				    return _G.GetItemCount(BREWFEST_PRIZE_TOKEN, true)
			    end
	},
	{
		item_id = LOVE_TOKEN,
		countFunc = function()
				    return _G.GetItemCount(LOVE_TOKEN, true)
			    end
	},
}

-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local startupTimer

-------------------------------------------------------------------------------
-- Helper functions
-------------------------------------------------------------------------------
local function getValue(info)
	return Broker_CurrencyCharDB[info[# info]]
end

local function setValue(info, value)
	Broker_CurrencyCharDB[info[# info]] = true and value or nil
	Broker_Currency:Update()
end

local function getColorValue(info)
	local color = Broker_CurrencyCharDB[info[# info]]
	return color.r, color.g, color.b, color.a
end

local function setColorValue(info, r, g, b, a)
	local color = Broker_CurrencyCharDB[info[# info]]
	color.r, color.g, color.b, color.a = r, g, b, a
	Broker_Currency:Update()
end

-- Data is saved per realm/character in Broker_CurrencyDB
-- Options are saved per character in Broker_CurrencyCharDB
-- There is separate settings for display of the broker, and the summary display on the tooltip
local sName, title, sNotes, enabled, loadable, reason, security = GetAddOnInfo("Broker_Currency")
local sName = GetAddOnMetadata("Broker_Currency", "X-BrokerName")

Broker_Currency.options = {
	type = "group",
	name = sName,
	get = getValue,
	set = setValue,
	childGroups = "tree",
	args = {}
}

do
	local settingsSliderIcon = ""

	local function setIconSize(info, value)
		Broker_CurrencyCharDB[info[# info]] = true and value or nil
		local iconSize = Broker_CurrencyCharDB.iconSize
		Broker_Currency.options.args.iconSize.name = string.format(settingsSliderIcon, 8, iconSize, iconSize)
		Broker_Currency:Update()
	end

	local function setIconSizeGold(info, value)
		Broker_CurrencyCharDB[info[# info]] = true and value or nil
		local iconSize = Broker_CurrencyCharDB.iconSizeGold
		Broker_Currency.options.args.iconSizeGold.name = string.format(_G.GOLD_AMOUNT_TEXTURE, 8, iconSize, iconSize)
		Broker_Currency:Update()
	end

	function Broker_Currency:InitializeOptions()
		-- Icons, names and textures for the currencies
		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local item_id = tokenInfo.item_id
			local currency_id = tokenInfo.currency_id

			if item_id then
				local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = _G.GetItemInfo(item_id)

				if itemTexture then
					tokenInfo.itemName = itemName
					tokenInfo.settingIcon = "\124T" .. itemTexture .. ":24:24\124t"
					tokenInfo.brokerIcon = DISPLAY_ICON_STRING1 .. itemTexture .. DISPLAY_ICON_STRING2
				end
			elseif currency_id then
				local name, _, icon_name = _G.GetCurrencyInfo(currency_id)

				if icon_name then
					tokenInfo.itemName = name
					tokenInfo.settingIcon = "\124T" .. "Interface\\Icons\\" .. icon_name .. ":24:24\124t"
					tokenInfo.brokerIcon = DISPLAY_ICON_STRING1 .. "Interface\\Icons\\" .. icon_name .. DISPLAY_ICON_STRING2
				end
			end
		end

		settingsSliderIcon = DISPLAY_ICON_STRING1 .. "Interface\\Icons\\" .. select(3, _G.GetCurrencyInfo(CONQUEST_POINTS)) .. DISPLAY_ICON_STRING2

		Broker_Currency.options.args = {
			header1 = {
				type = "description",
				order = 3,
				name = sNotes,
				cmdHidden = true
			},
			header2 = {
				type = "description",
				order = 5,
				name = sVersion,
				cmdHidden = true
			},
			iconSize = {
				type = "range",
				order = 10,
				name = string.format(settingsSliderIcon, 8, 16, 16),
				desc = _G.TOKENS,
				min = 1, max = 32, step = 1, bigStep = 1,
				set = setIconSize,
			},
			iconSizeGold = {
				type = "range",
				order = 10,
				name = string.format(_G.GOLD_AMOUNT_TEXTURE, 8, 16, 16),
				desc = _G.MONEY,
				min = 1, max = 32, step = 1, bigStep = 1,
				set = setIconSizeGold,
			},
			brokerDisplay = {
				type = "group",
				name = _G.DISPLAY,
				order = 20,
				inline = true,
				childGroups = "tree",
				args = {
					showGold = {
						type = "toggle",
						name = ICON_GOLD,
						desc = _G.MONEY,
						order = 1,
						width = "half",
					},
					showSilver = {
						type = "toggle",
						name = ICON_SILVER,
						desc = _G.MONEY,
						order = 2,
						width = "half",
					},
					showCopper = {
						type = "toggle",
						name = ICON_COPPER,
						desc = _G.MONEY,
						order = 3,
						width = "half",
					},
				},
			},
			statisticsDisplay = {
				type = "group",
				name = _G.STATISTICS,
				order = 30,
				inline = true,
				childGroups = "tree",
				args = {
					summaryPlayerSession = {
						type = "toggle",
						name = PLAYER_NAME,
						order = 1,
						width = "full",
					},
					summaryRealmToday = {
						type = "toggle",
						name = sToday,
						order = 2,
						width = "full",
					},
					summaryRealmYesterday = {
						type = "toggle",
						name = sYesterday,
						order = 3,
						width = "full",
					},
					summaryRealmThisWeek = {
						type = "toggle",
						name = sThisWeek,
						order = 4,
						width = "full",
					},
					summaryRealmLastWeek = {
						type = "toggle",
						name = sLastWeek,
						order = 5,
						width = "full",
					},
				},
			},
			summaryDisplay = {
				type = "group",
				name = _G.ACHIEVEMENT_SUMMARY_CATEGORY,
				order = 40,
				inline = true,
				childGroups = "tree",
				args = {
					summaryGold = {
						type = "toggle",
						name = ICON_GOLD,
						desc = _G.MONEY,
						order = 1,
						width = "half",
					},
					summarySilver = {
						type = "toggle",
						name = ICON_SILVER,
						desc = _G.MONEY,
						order = 2,
						width = "half",
					},
					summaryCopper = {
						type = "toggle",
						name = ICON_COPPER,
						desc = _G.MONEY,
						order = 3,
						width = "half",
					},
				},
			},
			color = {
				type = "group",
				name = _G.COLOR,
				order = 50,
				inline = true,
				childGroups = "tree",
				args = {
					summaryColorDark = {
						type = "color",
						name = _G.BACKGROUND,
						order = 11,
						get = getColorValue,
						set = setColorValue,
						hasAlpha = true,
					},
					summaryColorLight = {
						type = "color",
						name = _G.HIGHLIGHTING,
						order = 12,
						get = getColorValue,
						set = setColorValue,
						hasAlpha = true,
					},
				},
			},
			deleteCharacter = {
				type = "group",
				name = _G.DELETE,
				order = 60,
				inline = true,
				childGroups = "tree",
				args = {
				},
			},
		}
	end
end	-- do-block

local function GetKey(idnum, broker)
	if broker then
		return "show" .. idnum
	else
		return "summary" .. idnum
	end
end

-- Provide settings options for tokenInfo
local function SetOptions(brokerArgs, summaryArgs, tokenInfo, index)
	if tokenInfo.settingIcon then
		local idnum = tokenInfo.item_id or tokenInfo.currency_id
		local brokerName = GetKey(idnum, true)
		local summaryName = GetKey(idnum, nil)

		brokerArgs[brokerName] = {
			type = "toggle",
			order = index,
			name = tokenInfo.settingIcon,
			desc = tokenInfo.itemName,
			width = "half",
			get = function()
				      local key = brokerName
				      return Broker_CurrencyCharDB[key]
			      end,
			set = function(_, value)
				      local key = brokerName
				      Broker_CurrencyCharDB[key] = true and value or nil
				      Broker_Currency:Update()
			      end,
		}
		summaryArgs[summaryName] = {
			type = "toggle",
			order = index,
			name = tokenInfo.settingIcon,
			desc = tokenInfo.itemName,
			width = "half",
			get = function()
				      local key = summaryName
				      return Broker_CurrencyCharDB[key]
			      end,
			set = function(_, value)
				      local key = summaryName
				      Broker_CurrencyCharDB[key] = true and value or nil
				      Broker_Currency:Update()
			      end,
		}
	end
end

local function DeletePlayer(info)
	local PLAYER_NAME = info[# info]
	local deleteOptions = Broker_Currency.options.args.deleteCharacter.args

	deleteOptions[PLAYER_NAME] = nil
	deleteOptions[PLAYER_NAME .. "Name"] = nil
	deleteOptions[PLAYER_NAME .. "Spacer"] = nil
	Broker_CurrencyDB.realm[REALM_NAME][PLAYER_NAME] = nil
end

-- Provide settings options for tokenInfo
local function DeleteOptions(PLAYER_NAME, player_infoList, index)
	local deleteOptions = Broker_Currency.options.args.deleteCharacter.args

	if not deleteOptions[PLAYER_NAME] then
		deleteOptions[PLAYER_NAME .. "Name"] = {
			type = "description",
			order = index * 3,
			name = PLAYER_NAME,
		}

		deleteOptions[PLAYER_NAME] = {
			type = "execute",
			order = index * 3 + 1,
			name = _G.DELETE,
			func = DeletePlayer,
		}

		deleteOptions[PLAYER_NAME .. "Spacer"] = {
			type = "header",
			order = index * 3 + 2,
			name = "",
		}
	end
end

local AceCfgReg = LibStub("AceConfigRegistry-3.0")
local AceCfg = LibStub("AceConfig-3.0")
local brokerOptions = AceCfgReg:GetOptionsTable("Broker", "dialog", "LibDataBroker-1.1")

if not brokerOptions then
	brokerOptions = {
		type = "group",
		name = "Broker",
		args = {
		}
	}
	AceCfg:RegisterOptionsTable("Broker", brokerOptions)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker", "Broker")
end

AceCfg:RegisterOptionsTable("Broker_Currency", Broker_Currency.options)
Broker_Currency.menu = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Broker_Currency", sName, "Broker")

local tooltipBackdrop = {
	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
	tile = false,
	tileSize = 16,
	insets = {
		left = 0,
		right = 0,
		top = 2,
		bottom = 2
	},
}
local tooltipLines = {}
local tooltipLinesRecycle = {}
local tooltipAlignment = {}
local tooltipHeader = {}
local temp = {}

function Broker_Currency:ShowTooltip(button)
	Broker_Currency:Update()
	local maxColumns = 0
	for index, rowList in pairs(tooltipLines) do
		local columns = 0
		for i in pairs(rowList) do
			columns = columns + 1
		end
		maxColumns = math.max(maxColumns, columns)
	end

	if maxColumns > 0 then
		tooltipAlignment[1] = "LEFT"

		for index = 2, maxColumns + 1 do
			tooltipAlignment[index] = "RIGHT"
		end

		for index = # tooltipAlignment, maxColumns + 2, -1 do
			tooltipAlignment[index] = nil
		end

		table.wipe(tooltipHeader)
		tooltipHeader[1] = " "

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			if tokenInfo.brokerIcon then
				local idnum = tokenInfo.item_id or tokenInfo.currency_id
				local key = GetKey(idnum, false)

				if Broker_CurrencyCharDB[key] then
					tooltipHeader[# tooltipHeader + 1] = tokenInfo.settingIcon
				end
			end
		end

		if Broker_CurrencyCharDB.summaryGold then
			tooltipHeader[# tooltipHeader + 1] = ICON_GOLD
		end

		if Broker_CurrencyCharDB.summarySilver then
			tooltipHeader[# tooltipHeader + 1] = ICON_SILVER
		end

		if Broker_CurrencyCharDB.summaryCopper then
			tooltipHeader[# tooltipHeader + 1] = ICON_COPPER
		end

		local tooltip = LibQTip:Acquire("Broker_CurrencyTooltip", maxColumns, unpack(tooltipAlignment))
		self.tooltip = tooltip

		local fontName, fontHeight, fontFlags = tooltip.regularFont:GetFont()

		fontPlus:SetFont(fontName, fontHeight, fontFlags)
		fontPlus:SetTextColor(0, 1, 0)
		fontPlus:SetJustifyH("RIGHT")
		fontPlus:SetJustifyV("MIDDLE")

		fontMinus:SetFont(fontName, fontHeight, fontFlags)
		fontMinus:SetTextColor(1, 0, 0)
		fontMinus:SetJustifyH("RIGHT")
		fontMinus:SetJustifyV("MIDDLE")

		fontWhite:SetFont(fontName, fontHeight, fontFlags)
		fontWhite:SetTextColor(1, 1, 1)
		fontWhite:SetJustifyH("RIGHT")
		fontWhite:SetJustifyV("MIDDLE")

		fontName, fontHeight, fontFlags = tooltip.headerFont:GetFont()
		fontLabel:SetFont(fontName, fontHeight, fontFlags)
		fontLabel:SetTextColor(1, 1, 0.5)
		fontLabel:SetJustifyH("LEFT")
		fontLabel:SetJustifyV("MIDDLE")

		tooltip:AddHeader(unpack(tooltipHeader))
		tooltip:SetFont(fontWhite)

		for index, rowList in pairs(tooltipLines) do
			local label = rowList[1]
			local currentRow = index + 1

			if label == " " then
				tooltip:AddSeparator()
			elseif label == _G.STATISTICS or label == sToday or label == sYesterday or label == sThisWeek or label == sLastWeek then
				tooltip:AddHeader(unpack(tooltipHeader))
				tooltip:SetCell(currentRow, 1, label, fontLabel)
			else
				tooltip:AddLine(unpack(rowList))

				if label == sPlus then
					for i, value in ipairs(rowList) do
						if value == 0 then
							tooltip:SetCell(currentRow, i, " ", fontPlus)
						else
							tooltip:SetCell(currentRow, i, value, fontPlus)
						end
					end
				elseif label == sMinus then
					for i, value in ipairs(rowList) do
						if value == 0 then
							tooltip:SetCell(currentRow, i, " ", fontMinus)
						else
							tooltip:SetCell(currentRow, i, value, fontMinus)
						end
					end
				elseif label == sTotal then
					for i, value in ipairs(rowList) do
						if value and type(value) == "number" then
							if value == 0 then
								tooltip:SetCell(currentRow, i, " ", fontMinus)
							elseif value < 0 then
								tooltip:SetCell(currentRow, i, -1 * value, fontMinus)
							else
								tooltip:SetCell(currentRow, i, value, fontPlus)
							end
						end
					end
				end
				tooltip:SetCell(currentRow, 1, label, fontLabel)
			end
		end

		-- Color the even columns
		local column
		local summaryColorLight = Broker_CurrencyCharDB.summaryColorLight

		if summaryColorLight.a > 0 then
			for index = 2, maxColumns, 2 do
				tooltip:SetColumnColor(index, summaryColorLight.r, summaryColorLight.g, summaryColorLight.b, summaryColorLight.a)
			end
		end
		local summaryColorDark = Broker_CurrencyCharDB.summaryColorDark

		if _G.TipTac and _G.TipTac.AddModifiedTip then
			-- Pass true as second parameter because hooking OnHide causes C stack overflows
			_G.TipTac:AddModifiedTip(tooltip, true)
		elseif summaryColorDark.a > 0 then
			tooltip:SetBackdrop(tooltipBackdrop)
			tooltip:SetBackdropColor(summaryColorDark.r, summaryColorDark.g, summaryColorDark.b, summaryColorDark.a)
		end
		tooltip:SmartAnchorTo(button)
		tooltip:Show()
	end
end


function Broker_Currency:AddLine(label, currencyList)
	local newIndex = # tooltipLines + 1

	if not tooltipLinesRecycle[newIndex] then
		tooltipLinesRecycle[newIndex] = {}
	end
	tooltipLines[newIndex] = tooltipLinesRecycle[newIndex]

	local line = tooltipLines[newIndex]
	table.wipe(line)

	line[1] = label

	if currencyList then
		-- Create Strings for the various currencies
		for index, tokenInfo in pairs(CURRENCY_DATA) do
			if tokenInfo.brokerIcon then
				local idnum = tokenInfo.item_id or tokenInfo.currency_id
				local key = GetKey(idnum, false)
				local count = currencyList[idnum] or 0

				if Broker_CurrencyCharDB[key] then
					if count ~= 0 then
						line[# line + 1] = count
					else
						line[# line + 1] = " "
					end
				end
			end
		end

		-- Create Strings for gold, silver, copper
		local money = currencyList.money or 0
		local moneySign = (money < 0) and -1 or 1
		money = money * moneySign

		local copper = money % 100
		money = (money - copper) / 100

		local silver = money % 100
		local gold = math.floor(money / 100)

		gold = gold * moneySign
		silver = silver * moneySign
		copper = copper * moneySign

		if gold + silver + copper ~= 0 then
			if Broker_CurrencyCharDB.summaryGold then
				line[# line + 1] = gold
			end

			if Broker_CurrencyCharDB.summarySilver then
				line[# line + 1] = silver
			end

			if Broker_CurrencyCharDB.summaryCopper then
				line[# line + 1] = copper
			end
		end
	end
end


do
	local offset

	function Broker_Currency:GetServerOffset()
		if offset then
			return offset
		end
		local serverHour, serverMinute = _G.GetGameTime()
		local utcHour = tonumber(date("!%H"))
		local utcMinute = tonumber(date("!%M"))
		local ser = serverHour + serverMinute / 60
		local utc = utcHour + utcMinute / 60

		offset = math.floor((ser - utc) * 2 + 0.5) / 2

		if offset >= 12 then
			offset = offset - 24
		elseif offset < -12 then
			offset = offset + 24
		end
		return offset
	end
end

local function GetToday(self)
	return math.floor((time() / 60 / 60 + self:GetServerOffset()) / 24)
end

local CreateMoneyString
do
	local concatList = {}
	-- Create the display string for a single line
	-- money is the gold.silver.copper amount
	-- broker is true if it is the broker string, nil if it is the tooltip summary string
	-- currencyList contains totals for the set of currencies
	function CreateMoneyString(currencyList)
		local money = currencyList.money

		-- Create Strings for the various currencies
		table.wipe(concatList)

		if currencyList then
			for index, tokenInfo in pairs(CURRENCY_DATA) do
				if tokenInfo.brokerIcon then
					local idnum = tokenInfo.item_id or tokenInfo.currency_id
					local key = GetKey(idnum, true)
					local count = currencyList[idnum] or 0
					local size = Broker_CurrencyCharDB.iconSize

					if count > 0 and Broker_CurrencyCharDB[key] then
						concatList[# concatList + 1] = string.format(tokenInfo.brokerIcon, count, size, size)
						concatList[# concatList + 1] = "  "
					end
				end
			end
		end

		-- Create Strings for gold, silver, copper
		local copper = money % 100
		money = (money - copper) / 100

		local silver = money % 100
		local gold = math.floor(money / 100)

		if gold > 0 and Broker_CurrencyCharDB.showGold then
			concatList[# concatList + 1] = string.format(_G.GOLD_AMOUNT_TEXTURE, gold, Broker_CurrencyCharDB.iconSizeGold, Broker_CurrencyCharDB.iconSizeGold)
			concatList[# concatList + 1] = " "
		end

		if (gold + silver > 0) and Broker_CurrencyCharDB.showSilver then
			concatList[# concatList + 1] = string.format(_G.SILVER_AMOUNT_TEXTURE, silver, Broker_CurrencyCharDB.iconSizeGold, Broker_CurrencyCharDB.iconSizeGold)
			concatList[# concatList + 1] = " "
		end

		if (gold + silver + copper > 0) and Broker_CurrencyCharDB.showCopper then
			concatList[# concatList + 1] = string.format(_G.COPPER_AMOUNT_TEXTURE, copper, Broker_CurrencyCharDB.iconSizeGold, Broker_CurrencyCharDB.iconSizeGold)
			concatList[# concatList + 1] = " "
		end

		return table.concat(concatList)
	end
end

function Broker_Currency:Update(event)		--, ...)
	if event == "PLAYER_ENTERING_WORLD" then
		Broker_Currency:RegisterEvents()
	end

	if event == "PLAYER_LEAVING_WORLD" then
		Broker_Currency:UnregisterEvents()
		return
	end

	if Broker_Currency.InitializeSettings then
		Broker_Currency.InitializeSettings()
	end

	if event == "PLAYER_REGEN_ENABLED" then
		Broker_Currency:RegisterEvent("BAG_UPDATE", "Update")
	end

	if event == "PLAYER_REGEN_DISABLED" then
		Broker_Currency:UnregisterEvent("BAG_UPDATE")
		return
	end

	if _G.GetItemCount(HEARTHSTONE_IDNUM) < 1 and _G.GetMoney() == 0 then
		-- ToDo: check all items for nil?
		return
	end

	local realmInfo = Broker_CurrencyDB.realmInfo[REALM_NAME]
	local player_info = Broker_CurrencyDB.realm[REALM_NAME][PLAYER_NAME]
	local current_money = _G.GetMoney()

	-- Update the current player info
	player_info.money = current_money

	-- Update Statistics
	local today = GetToday(self)

	if not self.lastTime then
		self.lastTime = today
	end
	local cutoffDay = today - 14

	if today > self.lastTime then
		player_info.gained[cutoffDay] = nil
		player_info.spent[cutoffDay] = nil
		realmInfo.gained[cutoffDay] = nil
		realmInfo.spent[cutoffDay] = nil
		player_info.gained[today] = player_info.gained[today] or {money = 0}
		player_info.spent[today] = player_info.spent[today] or {money = 0}
		realmInfo.gained[today] = realmInfo.gained[today] or {money = 0}
		realmInfo.spent[today] = realmInfo.spent[today] or {money = 0}
		self.lastTime = today
	end

	-- Update Money
	if self.last.money < current_money then
		self.gained.money = (self.gained.money or 0) + current_money - self.last.money
		player_info.gained[today].money = (player_info.gained[today].money or 0) + current_money - self.last.money
		realmInfo.gained[today].money = (realmInfo.gained[today].money or 0) + current_money - self.last.money
	elseif self.last.money > current_money then
		self.spent.money = (self.spent.money or 0) + self.last.money - current_money
		player_info.spent[today].money = (player_info.spent[today].money or 0) + self.last.money - current_money
		realmInfo.spent[today].money = (realmInfo.spent[today].money or 0) + self.last.money - current_money
	end
	self.last.money = current_money

	-- Update Tokens
	for index, tokenInfo in pairs(CURRENCY_DATA) do
		if tokenInfo.brokerIcon then
			local idnum = tokenInfo.item_id or tokenInfo.currency_id
			local count = tokenInfo.countFunc()

			player_info[idnum] = count

			local last_count = self.last[idnum]

			if last_count then
				if last_count < count then
					self.gained[idnum] = (self.gained[idnum] or 0) + count - last_count
					player_info.gained[today][idnum] = (player_info.gained[today][idnum] or 0) + count - last_count
					realmInfo.gained[today][idnum] = (realmInfo.gained[today][idnum] or 0) + count - last_count
				elseif last_count > count then
					self.spent[idnum] = (self.spent[idnum] or 0) + last_count - count
					player_info.spent[today][idnum] = (player_info.spent[today][idnum] or 0) + last_count - count
					realmInfo.spent[today][idnum] = (realmInfo.spent[today][idnum] or 0) + last_count - count
				end
			end
			self.last[idnum] = count
		end
	end

	-- Display the money string according to the broker settings
	Broker_Currency.ldb.text = CreateMoneyString(player_info)

	self.savedTime = time()
end


-- Add counts from player_info to totalList according to the summary settings this character is interested in
local function TotalCurrencies(totalList, player_info)
	for summaryName in pairs(Broker_CurrencyCharDB) do
		local countKey = tonumber(string.match(summaryName, "summary(%d+)"))
		local count = player_info[countKey]

		if count then
			totalList[countKey] = (totalList[countKey] or 0) + count
		end
	end
	totalList.money = (totalList.money or 0) + (player_info.money or 0)
end
-- /dump Broker_CurrencyDB.realm["Proudmoore"]["Bliksem"]

local totalList = {}
local weekGained = {}
local weekSpent = {}
local lastWeekGained = {}
local lastWeekSpent = {}
local profit = {}
local sortMoneyList = {}

-- Sorting is in descending order of money
local function SortByMoneyDescending(a, b)
	if a.player_info.money and b.player_info.money then
		return a.player_info.money > b.player_info.money
	elseif a.player_info.money then
		return true
	elseif b.player_info.money then
		return false
	else
		return true
	end
end

local function GetSortedPlayerInfo()
	local index = 1

	for PLAYER_NAME, player_info in pairs(Broker_CurrencyDB.realm[REALM_NAME]) do
		if not sortMoneyList[index] then
			sortMoneyList[index] = {}
		end
		sortMoneyList[index].PLAYER_NAME = PLAYER_NAME
		sortMoneyList[index].player_info = player_info
		index = index + 1
	end

	for i = # sortMoneyList, index, -1 do
		sortMoneyList[i] = nil
	end
	table.sort(sortMoneyList, SortByMoneyDescending)

	return sortMoneyList
end

-- Handle mouse enter event in our button
local function OnEnter(button)
	if startupTimer then
		return
	end

	if Broker_Currency.InitializeSettings then
		Broker_Currency.InitializeSettings()
	end
	table.wipe(tooltipLines)

	-- Display the money string according to the summary settings
	table.wipe(totalList)

	local sortedPlayerInfo = GetSortedPlayerInfo()

	for i, data in ipairs(sortedPlayerInfo) do
		Broker_Currency:AddLine(string.format("%s: ", data.PLAYER_NAME), data.player_info, fontWhite)
		TotalCurrencies(totalList, data.player_info)
	end

	-- Statistics
	local charDB = Broker_CurrencyCharDB

	if charDB.summaryPlayerSession or charDB.summaryRealmToday or charDB.summaryRealmYesterday or charDB.summaryRealmThisWeek or charDB.summaryRealmLastWeek then
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(_G.STATISTICS)
	end

	-- Session totals
	local self = Broker_Currency
	local gained = self.gained
	local spent = self.spent

	if charDB.summaryPlayerSession then
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(PLAYER_NAME)

		table.wipe(profit)

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local idnum = tokenInfo.item_id or tokenInfo.currency_id

			if idnum then
				profit[idnum] = (gained[idnum] or 0) - (spent[idnum] or 0)
			end
		end
		Broker_Currency:AddLine(sPlus, gained)
		Broker_Currency:AddLine(sMinus, spent)
		Broker_Currency:AddLine(sTotal, profit)
	end

	-- Today totals
	local realmInfo = Broker_CurrencyDB.realmInfo[REALM_NAME]
	gained = realmInfo.gained
	spent = realmInfo.spent

	if charDB.summaryRealmToday then
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(sToday)
		Broker_Currency:AddLine(" ")

		table.wipe(profit)

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local idnum = tokenInfo.item_id or tokenInfo.currency_id

			if idnum then
				profit[idnum] = (gained[self.lastTime][idnum] or 0) - (spent[self.lastTime][idnum] or 0)
			end
		end
		Broker_Currency:AddLine(sPlus, gained[self.lastTime])
		Broker_Currency:AddLine(sMinus, spent[self.lastTime])
		Broker_Currency:AddLine(sTotal, profit)
	end

	-- Yesterday totals
	if charDB.summaryRealmYesterday then
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(sYesterday)
		Broker_Currency:AddLine(" ")

		local yesterday = self.lastTime - 1
		table.wipe(profit)

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local idnum = tokenInfo.item_id or tokenInfo.currency_id

			if idnum then
				profit[idnum] = (gained[yesterday][idnum] or 0) - (spent[yesterday][idnum] or 0)
			end
		end
		Broker_Currency:AddLine(sPlus, gained[yesterday])
		Broker_Currency:AddLine(sMinus, spent[yesterday])
		Broker_Currency:AddLine(sTotal, profit)
	end

	-- This Week totals
	if charDB.summaryRealmThisWeek then
		table.wipe(weekGained)
		table.wipe(weekSpent)
		table.wipe(profit)

		for i = self.lastTime - 6, self.lastTime do
			for index, tokenInfo in pairs(CURRENCY_DATA) do
				local idnum = tokenInfo.item_id or tokenInfo.currency_id

				if idnum then
					weekGained[idnum] = (weekGained[idnum] or 0) + (gained[i] and gained[i][idnum] or 0)
					weekSpent[idnum] = (weekSpent[idnum] or 0) + (spent[i] and spent[i][idnum] or 0)
				end
			end
		end

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local idnum = tokenInfo.item_id or tokenInfo.currency_id

			if idnum then
				profit[idnum] = weekGained[idnum] - weekSpent[idnum]
			end
		end
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(sThisWeek)
		Broker_Currency:AddLine(" ")

		Broker_Currency:AddLine(sPlus, weekGained)
		Broker_Currency:AddLine(sMinus, weekSpent)
		Broker_Currency:AddLine(sTotal, profit)
	end

	-- Last Week totals
	if charDB.summaryRealmLastWeek then
		table.wipe(lastWeekGained)
		table.wipe(lastWeekSpent)
		table.wipe(profit)

		for i = self.lastTime - 13, self.lastTime - 7 do
			for index, tokenInfo in pairs(CURRENCY_DATA) do
				local idnum = tokenInfo.item_id or tokenInfo.currency_id

				if idnum then
					lastWeekGained[idnum] = (lastWeekGained[idnum] or 0) + (gained[i] and gained[i][idnum] or 0)
					lastWeekSpent[idnum] = (lastWeekSpent[idnum] or 0) + (spent[i] and spent[i][idnum] or 0)
				end
			end
		end

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			local idnum = tokenInfo.item_id or tokenInfo.currency_id

			if idnum then
				profit[idnum] = lastWeekGained[idnum] - lastWeekSpent[idnum]
			end
		end
		Broker_Currency:AddLine(" ")
		Broker_Currency:AddLine(sLastWeek)
		Broker_Currency:AddLine(" ")

		Broker_Currency:AddLine(sPlus, lastWeekGained)
		Broker_Currency:AddLine(sMinus, lastWeekSpent)
		Broker_Currency:AddLine(sTotal, profit)
	end

	-- Totals
	Broker_Currency:AddLine(" ")
	Broker_Currency:AddLine(_G.ACHIEVEMENT_SUMMARY_CATEGORY, totalList)

	Broker_Currency:ShowTooltip(button)
end


local function OnLeave()
	LibQTip:Release(Broker_Currency.tooltip)
	Broker_Currency.tooltip = nil
end


-- Set up as a LibBroker data source
Broker_Currency.ldb = LDB:NewDataObject("Broker Currency", {
						type = "data source",
						label = _G.CURRENCY,
						icon = "Interface\\MoneyFrame\\UI-GoldIcon",
						text = "Initializing...",
						OnClick = function(clickedframe, button)
								  if button == "RightButton" then
									  _G.InterfaceOptionsFrame_OpenToCategory(Broker_Currency.menu)
								  end
							  end,
						OnEnter = OnEnter,
						OnLeave = OnLeave,
					})


do
	local wtfDelay = 5 -- For stupid cases where Blizzard pretends a player has no loots, wait up to 15 seconds

	function Broker_Currency.InitializeSettings()
		-- No hearthstone and no money means trouble
		if startupTimer then
			Broker_Currency:CancelTimer(startupTimer)
			startupTimer = nil
		end

		if _G.GetItemCount(HEARTHSTONE_IDNUM) < 1 and _G.GetMoney() == 0 then
			if wtfDelay > 0 then
				startupTimer = Broker_Currency:ScheduleTimer(Broker_Currency.InitializeSettings, wtfDelay)
				wtfDelay = wtfDelay - 1
				return
			end
		end
		local self = Broker_Currency

		-- Set defaults
		if not Broker_CurrencyCharDB then
			Broker_CurrencyCharDB = {
				showCopper = true,
				showSilver = true,
				showGold = true,
				showToday = true,
				showYesterday = true,
				showLastWeek = true,
				summaryGold = true,
				summaryColorDark = {r = 0, g = 0, b = 0, a = 0},
				summaryColorLight = {r = 1, g = 1, b = 1, a = .3},
			}
		end

		self:InitializeOptions()

		if not Broker_CurrencyCharDB.iconSize then
			Broker_CurrencyCharDB.iconSize = 16
		end

		if not Broker_CurrencyCharDB.iconSizeGold then
			Broker_CurrencyCharDB.iconSizeGold = 16
		end

		if not Broker_CurrencyCharDB.summaryColorDark then
			Broker_CurrencyCharDB.summaryColorDark = {r = 0, g = 0, b = 0, a = 0}
		end

		if not Broker_CurrencyCharDB.summaryColorLight then
			Broker_CurrencyCharDB.summaryColorLight = {r = 1, g = 1, b = 1, a = .3}
		end

		if not Broker_CurrencyDB then
			Broker_CurrencyDB = {}
		end

		if not Broker_CurrencyDB.realm then
			Broker_CurrencyDB.realm = {}
		end

		if not Broker_CurrencyDB.realmInfo then
			Broker_CurrencyDB.realmInfo = {}
		end

		if not Broker_CurrencyDB.realmInfo[REALM_NAME] then
			Broker_CurrencyDB.realmInfo[REALM_NAME] = {}
		end

		if not Broker_CurrencyDB.realm[REALM_NAME] then
			Broker_CurrencyDB.realm[REALM_NAME] = {}
		end

		if not Broker_CurrencyDB.realm[REALM_NAME][PLAYER_NAME] then
			Broker_CurrencyDB.realm[REALM_NAME][PLAYER_NAME] = {}
		end

		local realmInfo = Broker_CurrencyDB.realmInfo[REALM_NAME]

		if not realmInfo.gained or type(realmInfo.gained) ~= "table" then
			realmInfo.gained = {}
		end

		if not realmInfo.spent or type(realmInfo.spent) ~= "table" then
			realmInfo.spent = {}
		end

		local player_info = Broker_CurrencyDB.realm[REALM_NAME][PLAYER_NAME]

		if not player_info.gained or type(player_info.gained) ~= "table" then
			player_info.gained = {}
		end

		if not player_info.spent or type(player_info.spent) ~= "table" then
			player_info.spent = {}
		end

		if not self.last then
			self.last = {}
		end
		local last = self.last

		for index, tokenInfo in pairs(CURRENCY_DATA) do
			if tokenInfo.brokerIcon then
				local idnum = tokenInfo.item_id or tokenInfo.currency_id
				local count = tokenInfo.countFunc()

				player_info[idnum] = count
				last[idnum] = count
			end
		end

		-- Initialize statistics
		self.last.money = _G.GetMoney()
		self.lastTime = GetToday(self)

		local lastWeek = self.lastTime - 13

		for day in pairs(player_info.gained) do
			if day < lastWeek then
				player_info.gained[day] = nil
			end
		end

		for day in pairs(player_info.spent) do
			if day < lastWeek then
				player_info.spent[day] = nil
			end
		end

		for day in pairs(realmInfo.gained) do
			if day < lastWeek then
				realmInfo.gained[day] = nil
			end
		end

		for day in pairs(realmInfo.spent) do
			if day < lastWeek then
				realmInfo.spent[day] = nil
			end
		end

		for i = self.lastTime - 13, self.lastTime do
			if not player_info.gained[i] or type(player_info.gained[i]) ~= "table" then
				player_info.gained[i] = {
					money = 0
				}
			end

			if not player_info.spent[i] or type(player_info.spent[i]) ~= "table" then
				player_info.spent[i] = {
					money = 0
				}
			end

			if not realmInfo.gained[i] or type(realmInfo.gained[i]) ~= "table" then
				realmInfo.gained[i] = {
					money = 0
				}
			end

			if not realmInfo.spent[i] or type(realmInfo.spent[i]) ~= "table" then
				realmInfo.spent[i] = {
					money = 0
				}
			end
		end
		self.gained = {
			money = 0
		}
		self.spent = {
			money = 0
		}
		self.sessionTime = time()
		self.savedTime = time()

		-- Add settings for the various currencies
		local brokerDisplay = Broker_Currency.options.args.brokerDisplay.args
		local summaryDisplay = Broker_Currency.options.args.summaryDisplay.args

		for index = 1, # CURRENCY_DATA, 1 do
			-- Offset by three to ensure that gold, silver, and copper come first in the list.
			SetOptions(brokerDisplay, summaryDisplay, CURRENCY_DATA[index], index + 3)
		end

		-- Add delete settings so deleted characters can be removed
		local index = 1

		for PLAYER_NAME in pairs(Broker_CurrencyDB.realm[REALM_NAME]) do
			DeleteOptions(PLAYER_NAME, Broker_CurrencyDB.realm[REALM_NAME], index)
			index = index + 1
		end
		Broker_Currency:UnregisterEvent("BAG_UPDATE")

		if startupTimer then
			Broker_Currency:CancelTimer(startupTimer)
			startupTimer = nil
		end

		-- Register for update events
		Broker_Currency:RegisterEvents()
		
		Broker_Currency:RegisterEvent("PLAYER_ENTERING_WORLD", "Update")
		Broker_Currency:RegisterEvent("PLAYER_LEAVING_WORLD", "Update")

		-- Done initializing
		Broker_Currency:SetScript("OnEvent", Broker_Currency.Update)
		Broker_Currency.InitializeSettings = nil

		self:Update()
	end
end	-- do-block

function Broker_Currency:RegisterEvents()
	Broker_Currency:RegisterEvent("HONOR_CURRENCY_UPDATE", "Update")
	Broker_Currency:RegisterEvent("MERCHANT_CLOSED", "Update")
	Broker_Currency:RegisterEvent("PLAYER_MONEY", "Update")
	Broker_Currency:RegisterEvent("PLAYER_TRADE_MONEY", "Update")
	Broker_Currency:RegisterEvent("TRADE_MONEY_CHANGED", "Update")
	Broker_Currency:RegisterEvent("SEND_MAIL_MONEY_CHANGED", "Update")
	Broker_Currency:RegisterEvent("SEND_MAIL_COD_CHANGED", "Update")

	Broker_Currency:RegisterEvent("PLAYER_REGEN_ENABLED", "Update")
	Broker_Currency:RegisterEvent("PLAYER_REGEN_DISABLED", "Update")
	Broker_Currency:RegisterEvent("BAG_UPDATE", "Update")
end

function Broker_Currency:UnregisterEvents()
	Broker_Currency:UnregisterEvent("HONOR_CURRENCY_UPDATE", "Update")
	Broker_Currency:UnregisterEvent("MERCHANT_CLOSED", "Update")
	Broker_Currency:UnregisterEvent("PLAYER_MONEY", "Update")
	Broker_Currency:UnregisterEvent("PLAYER_TRADE_MONEY", "Update")
	Broker_Currency:UnregisterEvent("TRADE_MONEY_CHANGED", "Update")
	Broker_Currency:UnregisterEvent("SEND_MAIL_MONEY_CHANGED", "Update")
	Broker_Currency:UnregisterEvent("SEND_MAIL_COD_CHANGED", "Update")

	Broker_Currency:UnregisterEvent("PLAYER_REGEN_ENABLED", "Update")
	Broker_Currency:UnregisterEvent("PLAYER_REGEN_DISABLED", "Update")
	Broker_Currency:UnregisterEvent("BAG_UPDATE", "Update")
end

function Broker_Currency:Startup(event, ...)
	if event == "BAG_UPDATE" then
		if startupTimer then
			Broker_Currency:CancelTimer(startupTimer)
		end
		startupTimer = Broker_Currency:ScheduleTimer(Broker_Currency.InitializeSettings, 4)
	end
end

-- Initialize after end of BAG_UPDATE events
Broker_Currency:RegisterEvent("BAG_UPDATE")
Broker_Currency:RegisterEvent("PLAYER_MONEY")
Broker_Currency:SetScript("OnEvent", Broker_Currency.Startup)

LibStub("AceTimer-3.0"):Embed(Broker_Currency)

-- This is only necessary if AddonLoader is present, using the Delayed load. -Torhal
if IsLoggedIn() then
	startupTimer = Broker_Currency:ScheduleTimer(Broker_Currency.InitializeSettings, 4)
end
